unit FlagDisplay;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs;//, umsglog;

type
  TFlagDisplay = class(TCustomControl)
  private
    fBmp: array[0..3]of TBitmap;
    fOnChange: TNotifyEvent;
    fState: cardinal;
    ElementDown: integer;
    fSize: integer;
    procedure CustomWndProc(var msg: TMessage);
    procedure PaintProgress(DC: HDC);
    procedure SetImageList(const Value: TImageList);
    procedure SetState(const Value: cardinal);
  public
    isTriState: boolean;
    procedure Toggle(index: integer);
    function GetState(index: integer): integer;  //0=off 1=on 2=nochange 3=toggle
    constructor Create(AOwner: TComponent); override;
    destructor destroy;override;
  published
    property State: cardinal read fState write SetState;
    property ImageList: TImageList write SetImageList;
    property OnChange: TNotifyEvent read fOnChange write fOnChange;
  end;

implementation

constructor TFlagDisplay.Create(AOwner: TComponent);
begin
  inherited;
  isTriState:= true;
  fState := $ffff0000;
  fSize:=16;
  Width:=fSize*8;
  Height:=fSize;
  fBmp[0]:=TBitmap.Create;
  fBmp[1]:=TBitmap.Create;
  fBmp[2]:=TBitmap.Create;
  fBmp[3]:=TBitmap.Create;
  if not (csDesigning in ComponentState) then WindowProc := CustomWndProc;
  BevelEdges := [beLeft, beTop, beRight, beBottom];
  BevelInner := bvLowered;
end;

destructor TFlagDisplay.destroy;
begin
  ImageList:=nil;
  fBmp[0].Free;
  fBmp[1].Free;
  fBmp[2].Free;
  fBmp[3].Free;
  inherited;
end;

procedure TFlagDisplay.PaintProgress(DC: HDC);
var muhDC: HDC; oldbmp: HBITMAP; i: integer;
begin
  if (fBmp[0]<>nil) and (fBmp[1]<>nil) then begin
    muhDC := CreateCompatibleDC(DC);
    oldbmp := SelectObject(muhDC,fbmp[0].Handle);
    for i:=0 to 7 do begin
      if GetState(i)=0 then BitBlt(DC,i*fSize,0, fSize, fSize, muhDC,0,0, SRCCOPY);
    end;
    SelectObject(muhDC,fbmp[1].Handle);
    for i:=0 to 7 do begin
      if GetState(i)=1 then BitBlt(DC,i*fSize,0, fSize, fSize, muhDC,0,0, SRCCOPY);
    end;
    SelectObject(muhDC,fbmp[2].Handle);
    for i:=0 to 7 do begin
      if GetState(i)=2 then BitBlt(DC,i*fSize,0, fSize, fSize, muhDC,0,0, SRCCOPY);
    end;
    SelectObject(muhDC,fbmp[3].Handle);
    for i:=0 to 7 do begin
      if GetState(i)=3 then
        BitBlt(DC,i*fSize,0, fSize, fSize, muhDC,0,0, SRCCOPY);
    end;
    SelectObject(muhDC,oldbmp);
    DeleteDC(muhDC);
  end;
end;

procedure TFlagDisplay.Toggle(index: integer);
var n: cardinal;
begin
  if isTriState then begin
    case (fState shr index) and $00010001 of
      0: n:=1;
      1: n:=$00010000;
      $00010000: n:=$00010001;
      else n:=0;
    end;
    fState := fState and not ($00010001 shl index) or (n shl index);
  end else begin
    fState := fState xor (1 shl index);
  end;
end;

procedure TFlagDisplay.CustomWndProc(var msg: TMessage);
var
  DC,MemDC: HDC;
  MemBitmap, OldBitmap: HBITMAP;
  PS: TPaintStruct;
  xPos: short;
begin
  case msg.Msg of
    WM_PAINT: begin
      DC := GetDC(0);
      MemBitmap := CreateCompatibleBitmap(DC, ClientRect.Right, ClientRect.Bottom);
      ReleaseDC(0, DC);
      MemDC := CreateCompatibleDC(0);
      OldBitmap := SelectObject(MemDC, MemBitmap);
      try
        DC := BeginPaint(Handle, PS);
        PaintProgress(MemDC);
        BitBlt(DC, 0, 0, ClientRect.Right, ClientRect.Bottom, MemDC, 0, 0, SRCCOPY);
        EndPaint(Handle, PS);
      finally
        SelectObject(MemDC, OldBitmap);
        DeleteDC(MemDC);
        DeleteObject(MemBitmap);
      end;
      msg.Result:=0;
      exit;
    end;
    WM_ERASEBKGND: begin
      msg.Result:=1;
    end;

    WM_LBUTTONDBLCLK,
    WM_LBUTTONDOWN: begin
      xPos := LOWORD(Msg.LParam);  // horizontal position of cursor
      ElementDown:= xPos div fSize;
      if (ElementDown>=0) and (ElementDown<8) then SetCapture(Handle);
    end;
    WM_LBUTTONUP: begin
      if (ElementDown>=0) and (ElementDown<8) then begin
        ReleaseCapture;
        Toggle(ElementDown);
        ElementDown:=-1;
        Invalidate;
        if Assigned(fOnChange)then fOnChange(self);
      end;
    end;

    else WndProc(msg);
  end;
end;

procedure TFlagDisplay.SetImageList(const Value: TImageList);
begin
  if Value=nil then exit;
  fSize := Value.Width;
  Width:=fSize*8;
  Height:=fSize;
  Value.GetBitmap(0,fBmp[0]);
  Value.GetBitmap(1,fBmp[1]);
  Value.GetBitmap(2,fBmp[2]);
  Value.GetBitmap(3,fBmp[3]);
end;

procedure TFlagDisplay.SetState(const Value: cardinal);
begin
  if fState = Value then exit;
  fState := Value;
  if Assigned(fOnChange)then fOnChange(self);
  Invalidate;
end;

function TFlagDisplay.GetState(index: integer): integer;
begin
  if isTriState then begin
    result := (fstate shr index) and $00010001;
    result := (result or (result shr 15)) and $3
  end else begin
    result := (fState shr index) and 1;
  end;
end;

end.
