unit realtimer;

interface
uses forms,ExtCtrls,windows,classes,math,Buttons,sysutils;
type
  TRealTimer = class(TThread) // the TRealTimer is an attempt to get realtime capabilities into windows
   private
    IsWaitableTimerEnabled,
    IsActive,
    fEnabled: boolean;
    fInterval: double;
    fOnTimer: TNotifyEvent;
    hTimer: THandle;
    fDateTime: TDateTime;
    procedure SetEnabled(const Value: boolean);
    procedure SetInterval(const Value: double);
    function GetTicks: double;
    procedure SetDateTime(const Value: TDateTime);
   protected
    procedure Execute;override;
   public
    CpuFreq: double;
    property Enabled: boolean read fEnabled write SetEnabled;
    property Interval: double read fInterval write SetInterval; // seconds
    property DateTime: TDateTime read fDateTime write SetDateTime;
    procedure Resume;
    property OnTimer: TNotifyEvent read fOnTimer write fOnTimer;
    constructor Create;
    destructor Destroy;override;
  end;

procedure ReaDTimeStampCounter(var n: int64);

implementation
var
  hK32dll: THandle;
  pSetWaitableTimer: function(hTimer: THandle; var lpDueTime: TLargeInteger;
              lPeriod: Longint; pfnCompletionRoutine: TFNTimerAPCRoutine;
              lpArgToCompletionRoutine: Pointer; fResume: BOOL): BOOL; stdcall;
  pCreateWaitableTimer: function(lpTimerAttributes: PSecurityAttributes;
              bManualReset: BOOL; lpTimerName: PChar): THandle; stdcall;


procedure ReaDTimeStampCounter(var n: int64);
asm
  mov ECX,eax
  db $0f,$31 //RDTSC
  mov [ecx],eax
  mov [ecx+4],edx
end;

{ TRealTimer }
constructor TRealTimer.Create;
var n1,n2: int64; t1: cardinal;
begin
  inherited Create(true);
//  hTimer:= 0;

  if @pCreateWaitableTimer=nil then hTimer:=0 else
  hTimer:= pCreateWaitableTimer(nil,true,nil); // timer of 100ns accuracy, not available under win95


  fEnabled := false;
  fInterval := 0;
  fOnTimer := nil;
//  Priority := tpTimeCritical;
  Priority := tpHighest;

{  try
    t1:=GetTickCount;
    ReaDTimeStampCounter(n1);
    Sleep(200);
    t1:=GetTickCount-t1;
    ReaDTimeStampCounter(n2);
    CpuFreq := 1000*(n2-n1)/t1;
  except
    CpuFreq := 0;
  end;}
end;

destructor TRealTimer.Destroy;
begin
  Enabled := false; // stop the timer
  Terminate;  // set terminated true
  Resume;  // exit the thread loop
  while IsActive do
    Sleep(50);
  if hTimer<>0 then CloseHandle(hTimer);
  inherited;
end;

procedure PrintLastError;
var lpMsgBuf: PChar;
begin
  FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER or FORMAT_MESSAGE_FROM_SYSTEM,nil,GetLastError,0,@lpMsgBuf,0,nil);
  Application.MessageBox(lpMsgBuf,'error',0);
  LocalFree(cardinal(lpMsgBuf));
end;


procedure TRealTimer.Execute;
var t: int64; d: double; n: integer;
begin
  while not Terminated do begin
    //Priority := tpTimeCritical;
    Priority := tpHighest;
    IsActive:= true;
    while fEnabled and ((fInterval>0)or(fDateTime>0)) do begin
      if fDateTime>0 then begin
        d:= 24*3600*(fDateTime - Now);
        if d<0 then t:=0
        else t := round(-10000000*d)
      end else begin
        d:=fInterval;
        t := round(-10000000 * d);
      end;

      if (d>0) and (t<>0) then begin
        if hTimer<>0 then begin
          IsWaitableTimerEnabled:=true;
          if not pSetWaitableTimer(hTimer,t,0,nil,nil,false) then begin // set a timer of 100ns accuracy (it's not that accurate at all)
            PrintLastError;
            d:=d*0.1;
          end;
          WaitForSingleObjectEx(hTimer,round(d*1100),true);
          IsWaitableTimerEnabled:=false;
        end else begin   //  <- filthy hack for old win95
          while d>1 do begin
            Sleep(1000);
            if not Enabled then
              d:=0
            else d:=d-1;
          end;
          Sleep(round(d*1000));
        end;
      end;
      if fEnabled and ((fInterval>0)or(fDateTime>0)) then begin
        if Assigned(fOnTimer) then fOnTimer(self);
      end;
    end;

    IsActive:= false;
    Priority := tpNormal;
    Suspended:=not Terminated;
    //suspend;
  end;
end;

function TRealTimer.GetTicks: double;
var n: int64;
begin
  if CpuFreq=0 then result := GetTickCount*0.001
  else begin
    ReaDTimeStampCounter(n);
    result:= n/CpuFreq;
  end;
end;


procedure TRealTimer.SetEnabled(const Value: boolean);
var t: int64;
begin
  if fEnabled = Value then exit;
  fEnabled := Value;
  if fEnabled then begin
    if ((fInterval>0)or(fDateTime>0)) and Suspended and not isActive then Resume;
  end else begin
    if IsWaitableTimerEnabled and not Suspended and IsActive then begin
      t:=-1;
      if hTimer<>0 then if pSetWaitableTimer(hTimer,t,0,nil,nil,false) then begin // signal the timer immediately
        while not Suspended and IsActive do begin // let it happen
          Application.ProcessMessages;
          Sleep(10);
        end;
      end;
    end;
  end;
end;

procedure TRealTimer.SetInterval(const Value: double);
var t: int64;
begin
  if (fDateTime=0)and(fInterval = Value) then exit;
  fInterval := Value;
  fDateTime := 0;
  if not fEnabled then exit;

  if fInterval>0 then begin
    if Suspended and not isActive then Resume;
  end else begin
    if IsWaitableTimerEnabled and not Suspended and IsActive then begin
      t:=-1;
      if hTimer<>0 then pSetWaitableTimer(hTimer,t,0,nil,nil,false); // signal the timer immediately
      while not Suspended and IsActive do begin // let it happen
        Sleep(10);
      end;
    end;
  end;
end;

procedure TRealTimer.SetDateTime(const Value: TDateTime);
var t: int64;
begin
  if fDateTime = Value then exit;
  fDateTime := Value;
  if not fEnabled then exit;

  if fDateTime>0 then begin
    if Suspended and not isActive then Resume;
  end else begin
    if IsWaitableTimerEnabled and not Suspended and IsActive then begin
      t:=-1;
      if hTimer<>0 then pSetWaitableTimer(hTimer,t,0,nil,nil,false); // signal the timer immediately
      while not Suspended and IsActive do begin // let it happen
        Sleep(10);
      end;
    end;
  end;
end;

procedure TRealTimer.Resume;
begin

  inherited;
end;

procedure DoInit;
begin
  hK32dll := LoadLibrary('kernel32.dll');
  if hK32dll<>0 then begin
    @pCreateWaitableTimer := GetProcAddress(hK32dll,'CreateWaitableTimerA');
    @pSetWaitableTimer := GetProcAddress(hK32dll,'SetWaitableTimer');
  end else begin
    @pCreateWaitableTimer:=nil;
    @pSetWaitableTimer:=nil;
  end;
end;

procedure DoFinalize;
begin
  if hK32dll<>0 then FreeLibrary(hK32dll);
end;

initialization
  DoInit;
finalization
  DoFinalize;
end.
