unit tTProg;

interface
uses windows,classes,math,sysutils,FlagDisplay,realtimer;

const
  SEC2DATETIME = (1/(24*3600));

type
  TrtStateChange = procedure(StateMask: cardinal)of object;

  // program stuff
  prtTimeProgramStep = ^TrtTimeProgramStep;
  TrtTimeProgramStep = record
    StartTime: double;
    Interval: double;
    icnt,
    Iterations: integer;
    index: integer;
    selected: boolean;
    FlagDisplay: TFlagDisplay;
    NextEvent: double;
    next,prev: prtTimeProgramStep
  end;
  TrtTimeProgramStepArray = array[0..0]of TrtTimeProgramStep;
  prtTimeProgramSteps = ^TrtTimeProgramStepArray;
  // --

  TrtProgram = class(TObject)
  private
    CurrentEvent,
    FirstEvent: prtTimeProgramStep;
    Timer: TRealTimer;
    procedure ProgramStep;
    procedure SortIntoChain(p: prtTimeProgramStep);
  public
    capacity,count: integer;
    Items: prtTimeProgramSteps;
    IsLooping,
    IsExecuting: boolean;
    CurrentStep: integer;
    OnProgramFinished: TNotifyEvent;
    OnStateChange: TrtStateChange;
    StartTime: double;
    procedure Prepare;
    procedure TimerEvent(Sender: TObject);
    function Add:prtTimeProgramStep;
    procedure Clear;
    function Insert(index:integer):prtTimeProgramStep;
    procedure Delete(index:integer);
    procedure Step;
    procedure Start;
    procedure Stop;
    procedure Reset;
    procedure Sort;
    constructor Create;
    destructor Destroy;override;
  end;

implementation

procedure TrtProgram.Step;
begin
  inc(CurrentStep);
  if CurrentStep >= count then CurrentStep:=0;
  if (count>0)and assigned(OnStateChange)then OnStateChange(Items[CurrentStep].FlagDisplay.State);
end;

procedure TrtProgram.ProgramStep;
var
  CurrentInterval: double;
  s: prtTimeProgramStep;
begin
  if (CurrentEvent=nil) then exit;

  // apply change
  if assigned(OnStateChange)then OnStateChange(CurrentEvent.FlagDisplay.State);
  CurrentStep := CurrentEvent.index;

  // prepare next step

  if (CurrentEvent.Interval > 0)and((CurrentEvent.Iterations=0)or(CurrentEvent.icnt<CurrentEvent.Iterations)) then begin
    inc(CurrentEvent.icnt);
    CurrentInterval:=(CurrentEvent.Interval*SEC2DATETIME);
    CurrentEvent.NextEvent := CurrentEvent.NextEvent + CurrentInterval;

    if (CurrentEvent.next<>nil)and(CurrentEvent.next.NextEvent<CurrentEvent.NextEvent) then begin
      s:=CurrentEvent;  // remove CurrentEvent from the chain
      CurrentEvent := CurrentEvent.next;
      if s=FirstEvent then FirstEvent := CurrentEvent;
      if s.prev<>nil then s.prev.next := s.next;
      if s.next<>nil then s.next.prev := s.prev;
      s.next := nil;
      s.prev := nil;
      SortIntoChain(s);
    end else begin
      //if CurrentEvent.next=nil then CurrentEvent := nil;
    end;

  end else if (CurrentEvent.StartTime>0)and(CurrentEvent.StartTime<=1) then begin // daily events
    CurrentEvent.NextEvent := CurrentEvent.NextEvent + 1.0; // next day
    if (CurrentEvent.next=nil)or(CurrentEvent.next.NextEvent>CurrentEvent.NextEvent) then begin
      //same event again next
    end else begin
      s:=CurrentEvent;
      CurrentEvent := CurrentEvent.next;
      if s.prev<>nil then s.prev.next := s.next;
      if s.next<>nil then s.next.prev := s.prev;
      SortIntoChain(s);
    end;
  end else begin
    CurrentEvent := CurrentEvent.next;
  end;

  if CurrentEvent=nil then begin
    if IsLooping then begin
      Prepare;
      CurrentEvent := FirstEvent;
    end else begin
      if Assigned(OnProgramFinished) then OnProgramFinished(self);
      IsExecuting := false;
      Timer.Enabled:=false;
      exit;
    end;
  end;
  if not IsExecuting then exit;
  Timer.DateTime := CurrentEvent.NextEvent;
end;

procedure TrtProgram.TimerEvent(Sender: TObject);
begin
  if IsExecuting then begin
    ProgramStep;
  end;
end;

procedure TrtProgram.Start;
begin
  if count=0 then exit;
  IsExecuting := true;
  Timer.Interval:=0;
  Timer.Enabled:=true;
  Prepare;
  CurrentEvent:=FirstEvent;
  if CurrentEvent<>nil then Timer.DateTime := CurrentEvent.NextEvent;
end;

procedure TrtProgram.Stop;
begin
  IsExecuting:=false;
  Timer.Enabled:=false;
end;

procedure TrtProgram.Reset;
begin
  CurrentStep:=0;
  IsExecuting:=false;
  Timer.Enabled:=false;
end;

function TrtProgram.Add:prtTimeProgramStep;
begin
  if count>=capacity then begin
    inc(capacity,64);
    ReallocMem(Items,capacity*sizeof(TrtTimeProgramStep));
  end;
  result:=@Items[count];
  result.StartTime :=0;
  result.Interval := 0;
  result.FlagDisplay := nil;
  result.index := count;
  result.NextEvent := 0;
  inc(count);
end;

procedure TrtProgram.Clear;
var i: integer;
begin
  if IsExecuting then begin
    Timer.Enabled:=false;
  end;
  for i:=0 to count-1 do begin
    items[i].FlagDisplay.Free;
  end;
  count:=0;
  capacity:=0;
  FreeMem(Items);
  Items:=nil;
  FirstEvent:=nil;
end;

constructor TrtProgram.Create;
begin
  count:=0;
  capacity:=0;
  Items:=nil;
  FirstEvent:=nil;
  Timer:= TRealTimer.Create;
  Timer.Enabled:=false;
  Timer.Interval:=0;
  Timer.OnTimer:= TimerEvent;
  IsLooping:=false;
  IsExecuting:= false;
  CurrentStep:=0;
end;

procedure TrtProgram.Delete(index:integer);
var i:integer;
begin
  if index<0 then exit;
  if index<count then begin
    items[index].FlagDisplay.ImageList := nil;
    items[index].FlagDisplay.Free;
    dec(count);
    if index=count then exit;
    for i:=index to count-1 do begin
      CopyMemory(@Items[i],@Items[i+1],sizeof(TrtTimeProgramStep));
    end;
  end;
end;

destructor TrtProgram.Destroy;
var i: integer;
begin
  for i:=0 to count-1 do begin
    items[i].FlagDisplay.Free;
  end;
  FreeMem(Items);
  Timer.Free;
end;

function TrtProgram.Insert(index:integer):prtTimeProgramStep;
var i:integer;
begin
  if (index>=count) or (index<0) then begin
    result:=Add;
    exit;
  end;

  if count>=capacity then begin
    inc(capacity,64);
    ReallocMem(Items,capacity*sizeof(TrtTimeProgramStep));
  end;

  for i:=count-1 downto index do begin
    CopyMemory(@Items[i+1],@Items[i],sizeof(TrtTimeProgramStep));
  end;

  result:=@Items[index];
  inc(count);
end;

procedure TrtProgram.SortIntoChain( p: prtTimeProgramStep);
var t:prtTimeProgramStep;
label testthis;
begin
  if FirstEvent=nil then begin
    FirstEvent:=p;
    FirstEvent.next:=nil;
    FirstEvent.prev:=nil;
    exit;
  end;

testthis:
  t:=FirstEvent;
  if t.NextEvent > p.NextEvent then begin
    FirstEvent := p;
    p.prev := nil;
    p.next := t;
    t.prev := p;
    exit;
  end;
  while (t.next<>nil)and(t.next.NextEvent < p.NextEvent) do begin
    if t=t.next then begin
      //goto testthis;
      t.next:=nil;
      break;
    end;
    t := t.next;
  end;
  p.next := t.next;
  p.prev := t;
  t.next := p;
  if p.next<>nil then p.next.prev := p;
end;

procedure TrtProgram.Prepare;
var i: integer;
begin
  if count=0 then exit;
  StartTime := Now;
  FirstEvent:= nil;
  for i:=0 to count-1 do begin
    items[i].icnt:= 0;
    if items[i].StartTime>0 then begin
      items[i].NextEvent:=items[i].StartTime;
      if items[i].NextEvent <= 1 then begin
        items[i].NextEvent := items[i].NextEvent + floor(Now);
        if items[i].NextEvent < Now then items[i].NextEvent := items[i].NextEvent + 1.0;
      end;
    end else begin
      items[i].NextEvent:=  StartTime + items[i].StartTime * -SEC2DATETIME;
    end;
    if StartTime <= items[i].NextEvent then SortIntoChain(@items[i]);
  end;
end;

function SortTimerProgramItems(a,b: pointer): integer;
var aa,bb: prtTimeProgramStep; f: double;  d1,d2: double;
begin
  aa:=a;
  bb:=b;

  d1:=aa.StartTime;
  d2:=bb.StartTime;
  if d1>0 then d1:=d1 + 10000;
  if d2>0 then d2:=d2 + 10000;
  f := abs(d1) - abs(d2);
  if f>0 then result:=1 else
  if f<0 then result:=-1 else
  result:=aa.index - bb.index;
end;

procedure TrtProgram.Sort;
var L: TList;  i: integer;  oldItems: prtTimeProgramSteps;
begin
 if count<2 then exit;
  L:= TList.Create;
  for i:=0 to count-1 do begin
    L.Add(@Items[i]);
  end;
  L.Sort(SortTimerProgramItems);
  oldItems:=Items;
  GetMem(Items,capacity*sizeof(TrtTimeProgramStep));
  for i:=0 to count-1 do begin
    CopyMemory(@Items[i],L[i],sizeof(TrtTimeProgramStep));
  end;
  FreeMem(oldItems);
  L.Free;
end;

end.
